// This file is part of BUDEES, a software that allows fast simulation of bubble dynamics in electrolyzers and their interactions with electrical parameters. 
// 
// Copyright (C) 2024 <CNRS-ENS Rennes-Université de Rennes-CY Cergy Paris Université-Université Paris Saclay-CNAM Paris-ENS Paris-Saclay>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/

/* Global const */
double Fcst = 96485.332; // Faraday cst [A.s/mol]
double Rcst = 8.314463; // Constante gaz parfait [J/mol.K]

/// Static pre-computation
double eThick, RSite[2], Fcapilaire[2], FcapilaireSq[2], RFritzExtra[2], Rmini[2],
nRPrefactor, nFritz[2], RFritzMax, RFritzMin, buoyancyPrefactor, QnPrefactor[2], Tprefactor[2], growingKineticFractionWait, uAscentMin,
tafelJ0A0[2], bPrefactor,
interactionLongMaximumRangeSq, interactionLubMaximumRangeSq, interactionRepulsiveRangeSq, interactionCollisionRangeSq,
interactionSafetyHalfRange, interactionSafetyHalfRangeInv, interactionSafetyHalfRangeSq, interactionSafetyRange, interactionSafetyRangeSq;
int bubbleLifeCount;

void staticPreComputation() {
	// Diametre moyen
	eThick = (RFritz[0] + RFritz[1]);
	for (int i = 0; i < 2; i++) {
		// Taille du défaut 'r' associé au site de nucléation pour le rayon de Fritz considérée (tel que Fcap = Fbuoyancy)
		RSite[i] = 2*CUBE(RFritz[i])*rho*gravity/(3*gammaLiquidGas[i]);
		// Force capilaire sur un site (Fcap = 2 PI r gamma)
		Fcapilaire[i] = 2.0*M_PI*RSite[i]*gammaLiquidGas[i];
		FcapilaireSq[i] = SQ(Fcapilaire[i]);
		//...
		RFritzExtra[i] = (extraRFritz+1)*RFritz[i];
		Rmini[i] = detachementMinimalRadiusNorm*RFritz[i];
	}
	// Préfacteur pour le calcul de la quantité de matière dans une sphère (n = [4/3]PI * P0/RT * r^3)
	nRPrefactor = 4.0*M_PI*P0/(3.0*Rcst*T0);
	// Quantité de matière au détachement de Fritz
	for (int i = 0; i < 2; i++) {
		nFritz[i] = nRPrefactor*CUBE(RFritz[i]);
	}
	// ...
	RFritzMax = mathMax(RFritz[0], RFritz[1]);
	RFritzMin = mathMin(RFritz[0], RFritz[1]);
	// Préfacteur pour la force de flottaison Fbuoyancy = [4/3]PI*rho*g * r^3
	buoyancyPrefactor = (4.0/3.0)*M_PI*rho*gravity;
	for (int i = 0; i < 2; i++) {
		// Préfacteur pour le calcul de la charge (Jelec = [Nelec*F] * Jmolaire)
		QnPrefactor[i] = Nelectronic[i]*Fcst;
		// Préfacteur pour le calcul de la periode de nucleation
		Tprefactor[i] = Nelectronic[i]*Fcst*nFritz[i];
	}
	// ...
	growingKineticFractionWait = 1.0-growingKineticFraction;
	// Plus petite vitesse d'ascension naturelle (pour les plus petites bulles c-a-d les bulles à RFritz)
	uAscentMin = 2*rho*gravity*mathSQ(RFritzMin)/(9*mu);

	// Préfacteur de la dérivée de Tafel
	for (int i = 0; i < 2; i++) {
		tafelJ0A0[i] = tafelJ0[i]/tafelA0[i];
	}
	// Préfacteur du terme de flotaison
	bPrefactor = (2.0*rho*gravity)/(9.0*mu); // = (4/3)*PI*rho*g*R3 / 6*pi*mu*R

	// Répétition des bulles -> 2 vies
	bubbleLifeCount = (repeatYBubble ? 2 : 1);

	// Version caré et inverse des parametres numériques
	interactionLongMaximumRangeSq = mathSQ(interactionLongMaximumRange);
	interactionLubMaximumRangeSq = mathSQ(interactionLubMaximumRange);
	interactionRepulsiveRangeSq = mathSQ(interactionRepulsiveRange);
	interactionCollisionRangeSq = mathSQ(interactionCollisionRange);

	interactionSafetyHalfRange = interactionSafetyD+1.0;
	interactionSafetyHalfRangeInv = 1.0/interactionSafetyHalfRange;
	interactionSafetyHalfRangeSq = mathSQ(interactionSafetyHalfRange);
	interactionSafetyRange = 2.0+interactionSafetyD;
	interactionSafetyRangeSq = mathSQ(interactionSafetyRange);
}

#define TafelAnodeJ(deltaphi) (tafelJ0[ANODE]*exp((deltaphi)/tafelA0[ANODE]))
#define TafelCathodeJ(deltaphi) (-tafelJ0[CATHODE]*exp((deltaphi)/tafelA0[CATHODE]))

#define dTafelAnodeJ(deltaphi) (-tafelJ0A0[ANODE]*exp((deltaphi)/tafelA0[ANODE]))
#define dTafelCathodeJ(deltaphi) (-tafelJ0A0[CATHODE]*exp((deltaphi)/tafelA0[CATHODE]))

