// This file is part of BUDEES, a software that allows fast simulation of bubble dynamics in electrolyzers and their interactions with electrical parameters. 
// 
// Copyright (C) 2024 <CNRS-ENS Rennes-Université de Rennes-CY Cergy Paris Université-Université Paris Saclay-CNAM Paris-ENS Paris-Saclay>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/

// Calcul la carte des densités de courant sqrt(Jx2 + Jy2)
// J = sigma.grad(phi)
MatrixXd computeCurrent(Parameters& parameters, MatrixXi& MASK, MatrixTension& TENSION) {
	int* N = parameters.N;
	double dxInvSigma = sigma/parameters.dL[0], dyInvSigma = sigma/parameters.dL[1];
	MatrixXd CURRENT(N[0], N[1]);
	for (int i = 0; i < N[0]; i++) for (int j = 0; j < N[1]; j++) {
		// Electrolyte
		if (MASK(i,j) == 0) {
			double Jx = 0, Jy = 0;
			// dx
			if (i > 0) {
				// electrolyte
				if (MASK(i-1,j) == 0) Jx = dxInvSigma*(TENSION.VALUES(i,j) - TENSION.VALUES(i-1,j));
				// anode
				else if (MASK(i-1,j) == 1) Jx = TafelAnodeJ(TENSION.phi_anode-TENSION.VALUES(i,j));
				// cathode
				else if (MASK(i-1,j) == -1) Jx = TafelCathodeJ(TENSION.VALUES(i,j)-TENSION.phi_cathode);
			}
			// dy
			int jm = (j > 0) ? j-1 : (N[1]-1);
			// electrolyte
			if (MASK(i,jm) == 0) Jy = dyInvSigma*(TENSION.VALUES(i,j) - TENSION.VALUES(i,jm));
			// anode
			else if (MASK(i,jm) == 1) Jy = TafelAnodeJ(TENSION.phi_anode-TENSION.VALUES(i,j));
			// cathode
			else if (MASK(i,jm) == -1) Jy = TafelCathodeJ(TENSION.VALUES(i,j)-TENSION.phi_cathode);


			// norm
			CURRENT(i,j) = sqrt(SQ(Jx) + SQ(Jy));
		}
		// Insulation
		else if (MASK(i, j) == -2) {
			CURRENT(i, j) = -1;
			continue;
		}
		// Electrode (anode/cathode)
		else if ((MASK(i, j) == 1) || (MASK(i, j) == -1)) {
			CURRENT(i, j) = 0;
		}
	}
	return CURRENT;
}

// Calcul la quantité de courant arrivant en chaque noeud frontière d'une électrode
// I = S.Jtafel
MatrixXd computeSourceNormalized(Parameters& parameters, MatrixXi& MASK, MatrixTension& TENSION) {
	int* N = parameters.N;
	double Sx = eThick*parameters.dL[1], Sy = eThick*parameters.dL[0];
	MatrixXd SOURCES(N[0], N[1]);

	for (int i = 0; i < N[0]; i++) for (int j = 0; j < N[1]; j++) {
		double src = 0;
		if (MASK(i,j) == 0) {
			// i-1
			if (i > 0) {
				// anode
				if (MASK(i-1,j) == 1) src += Sx*TafelAnodeJ(TENSION.phi_anode-TENSION.VALUES(i,j));
				// cathode
				else if (MASK(i-1,j) == -1) src += Sx*TafelCathodeJ(TENSION.VALUES(i,j)-TENSION.phi_cathode);
			}

			// i+1
			if (i < (N[0]-1)) {
				// anode
				if (MASK(i+1,j) == 1) src += Sx*TafelAnodeJ(TENSION.phi_anode-TENSION.VALUES(i,j));
				// cathode
				else if (MASK(i+1,j) == -1) src += Sx*TafelCathodeJ(TENSION.VALUES(i,j)-TENSION.phi_cathode);
			}

			// j-1
			int jm = (j > 0) ? j-1 : (N[1]-1);
			// anode
			if (MASK(i,jm) == 1) src += Sy*TafelAnodeJ(TENSION.phi_anode-TENSION.VALUES(i,j));
			// cathode
			else if (MASK(i,jm) == -1) src += Sy*TafelCathodeJ(TENSION.VALUES(i,j)-TENSION.phi_cathode);

			// j+1
			int jp = (j < (N[1]-1)) ? j+1 : 0;
			// anode
			if (MASK(i,jp) == 1) src += Sy*TafelAnodeJ(TENSION.phi_anode-TENSION.VALUES(i,j));
			// cathode
			else if (MASK(i,jp) == -1) src += Sy*TafelCathodeJ(TENSION.VALUES(i,j)-TENSION.phi_cathode);
		}
		SOURCES(i,j) = src;
	}
	return SOURCES;
}

