// This file is part of BUDEES, a software that allows fast simulation of bubble dynamics in electrolyzers and their interactions with electrical parameters. 
// 
// Copyright (C) 2024 <CNRS-ENS Rennes-Université de Rennes-CY Cergy Paris Université-Université Paris Saclay-CNAM Paris-ENS Paris-Saclay>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/



/* -------- Compute bubble evolution --------- */
/* ------------------------------------------- */
#include "utils/funinter.h"
#include "utils/fundyn.h"
#include "utils/funvelo.h"
void computeBubble(Parameters& parameters, ListBubbles& Bubbles, vector<Wall>& Walls, ElectrolysisStatistics& Statistics, double t, double dt, bool* stopFlag) {
	// 0| Prepare variables
	Bubbles.maxRadius = RFritzMax; Bubbles.maxVelocity = 0;
	list<Bubble>::iterator it;
	for (it = Bubbles.attached.begin(); it != Bubbles.attached.end(); it++) {
		it->interactor.count = 0; // Reset bubble-bubble interactions
	}
	for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
		it->interactor.count = 0; // Reset bubble-bubble interactions
	}


	// 1-2| Determine interactions and build mobility matrix
	std::vector<Trip> M_coefs; // Matrix coefficients for velocity system to solve
	double Mij[4] { 0, 0, 0, 0 }; // Temp matrix coefficients

	int i = -1;
	for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
		i++;
		// 1| Test which wall and bubbles are supposed to interact significantly
		checkWalls(it->interactor.walls, Walls, it->x, it->y, 0.5*interactionLongMaximumRange*it->r);
		checkBubbles(it, i, Bubbles, parameters.L[1], stopFlag);

		// 2| Computing mobility matrix - line i
		setMobilityMatrixSphere_I(Mij);
		addMobilityMatrixPlane(Mij, &(*it));
		M_coefs.push_back(Trip(2*i+0, 2*i+0, Mij[0])); M_coefs.push_back(Trip(2*i+0, 2*i+1, Mij[1]));
		M_coefs.push_back(Trip(2*i+1, 2*i+0, Mij[2])); M_coefs.push_back(Trip(2*i+1, 2*i+1, Mij[3]));

		double sqI = SQ(it->r);
		for (int jj = 0; jj < it->interactor.count; jj++) {
			int j = it->interactor.interactionIndexs[jj];
			if (j == -1) continue;

			Bubble* jt = it->interactor.bubbles[jj];
			if (jt->r == 0) continue;

			double sqJ = SQ(jt->r);
			int jMode = it->interactor.imagedModes[jj];

			if (it->interactor.distancesSqNorm[jj] < interactionLongMaximumRangeSq) {
				setMobilityMatrixSphere_J(Mij, 0.75*it->r, sqI, sqJ, it->x-jt->x, it->y-yRepeat(jt->y, parameters.L[1], jMode));
				M_coefs.push_back(Trip(2*i+0, 2*j+0, Mij[0])); M_coefs.push_back(Trip(2*i+0, 2*j+1, Mij[1]));
				M_coefs.push_back(Trip(2*i+1, 2*j+0, Mij[2])); M_coefs.push_back(Trip(2*i+1, 2*j+1, Mij[3]));
			}
		}
	}

	// 3| Invert mobility matrix
	SpMat Minv;
	if (Bubbles.countDetached > 0) {
		SpMat M(2*Bubbles.countDetached, 2*Bubbles.countDetached);
		M.setFromTriplets(M_coefs.begin(), M_coefs.end());

		if (invertLU(Minv, M) != 0) {
			// Decomposition failed
			errorProgress("mobility matrix inverting error @t=" + std::to_string(t) + "s");
			*stopFlag = true;
			return;
		}
	}

	// 4| Build resistance matrix and source term
	std::vector<Trip> R_coefs; // Matrix coefficients for velocity system to solve
	double Rij[4] { 0, 0, 0, 0 }; // Temp matrix coefficients
	Eigen::VectorXd U0(2*Bubbles.countDetached); // Source term

	i = -1;
	for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
		i++;
		// Build source term
		U0[2*i+0] = 0;
		U0[2*i+1] = -bPrefactor*SQ(it->r);

		// Add repulsive term
		addSphereSphereRepulsive(U0, &(*it), i);

		// Add inverted mobility matrix
		Rij[0] = -Minv.coeff(2*i+0, 2*i+0); Rij[1] = -Minv.coeff(2*i+0, 2*i+1);
		Rij[2] = -Minv.coeff(2*i+1, 2*i+0); Rij[3] = -Minv.coeff(2*i+1, 2*i+1);

		// Add sphere-sphere lubrification
		addSphereSphereContact_I(Rij, parameters.L[1], &(*it));
		// Add sphere-plan lubrification
		addSpherePlanContact_I(Rij, &(*it));

		// Build resistance matrix ii
		R_coefs.push_back(Trip(2*i+0, 2*i+0, Rij[0])); R_coefs.push_back(Trip(2*i+0, 2*i+1, Rij[1]));
		R_coefs.push_back(Trip(2*i+1, 2*i+0, Rij[2])); R_coefs.push_back(Trip(2*i+1, 2*i+1, Rij[3]));

		for (int jj = 0; jj < it->interactor.count; jj++) {
			int j = it->interactor.interactionIndexs[jj];
			if (j == -1) continue;

			Bubble* jt = it->interactor.bubbles[jj];
			if (jt->r == 0) continue;

			int jMode = it->interactor.imagedModes[jj];

			if (it->interactor.distancesSqNorm[jj] < interactionLongMaximumRangeSq) {
				// Add inverted mobility matrix
				double ratio_ji = jt->r/it->r;
				Rij[0] = -ratio_ji*Minv.coeff(2*i+0, 2*j+0); Rij[1] = -ratio_ji*Minv.coeff(2*i+0, 2*j+1);
				Rij[2] = -ratio_ji*Minv.coeff(2*i+1, 2*j+0); Rij[3] = -ratio_ji*Minv.coeff(2*i+1, 2*j+1);

				// Add sphere-sphere lubrification
				if (it->interactor.distancesSqNorm[jj] < interactionLubMaximumRangeSq) {
					addSphereSphereContact_J(Rij, parameters.L[1], &(*it), jt, j, jMode);
				}

				// Build resistance matrix ij
				R_coefs.push_back(Trip(2*i+0, 2*j+0, Rij[0])); R_coefs.push_back(Trip(2*i+0, 2*j+1, Rij[1]));
				R_coefs.push_back(Trip(2*i+1, 2*j+0, Rij[2])); R_coefs.push_back(Trip(2*i+1, 2*j+1, Rij[3]));
			}
		}
	}

	// 5| Solve velocity
	Eigen::VectorXd U;
	if (Bubbles.countDetached > 0) {
		SpMat R(2*Bubbles.countDetached, 2*Bubbles.countDetached);
		R.setFromTriplets(R_coefs.begin(), R_coefs.end());
		Eigen::SparseLU<SpMat> solver;
		solver.compute(R);
		if (solver.info() != Eigen::Success) {
			// Decomposition failed
			errorProgress("resistance matrix solving error @t=" + std::to_string(t) + "s");
			*stopFlag = true;
			return;
		}
		U = solver.solve(U0);
	}

	// 6| Update velocity of detached bubbles
	i = -1;
	for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
		i++;
		it->ux = U(2*i+0);
		it->uy = U(2*i+1);
	}

	// 7| Check attached bubble detachment
	double Fdyn[2];
	for (it = Bubbles.attached.begin(); it != Bubbles.attached.end(); it++) {
		// Si le détachement dynamique est possible
		if (it->r != 0) {
			// Calcul des actions hydrodynamiques
			Fdyn[0] = 0; Fdyn[1] = 0;
			for (int kk = 0; kk < it->interactor.count; kk++) {
				Bubble *kt = it->interactor.bubbles[kk];
				int kMode = it->interactor.imagedModes[kk];
				if (it->interactor.distancesSqNorm[kk] < interactionCollisionRangeSq) {
					double deltaX = (it->x - kt->x), deltaY = (it->y - yRepeat(kt->y, parameters.L[1], kMode));
					double norm = kt->ux*deltaX + kt->uy*deltaY;
					if (norm > 0) {
						// Si contact avec une autre bulle alors elle détache immédiatement
						it->statLife = 0;
						break;
					}
				} else if (it->interactor.distancesSqNorm[kk] < interactionLubMaximumRangeSq) {
					computeSphereSphereForce(Fdyn, &(*it), kt, kMode, parameters.L[1]);
				}
			}
			// Ajout de la pesanteur et retrait de la composante reprise par la paroi
			Fdyn[1] += buoyancyPrefactor*CUBE(it->r);
			double proj = Fdyn[0]*it->site->nx + Fdyn[1]*it->site->ny;
			if (proj < 0) {
				Fdyn[0] -= proj*it->site->nx;
				Fdyn[1] -= proj*it->site->ny;
			}
			// Test de détachement hydrodynamique
			if ((SQ(Fdyn[0]) + SQ(Fdyn[1]) > FcapilaireSq[it->site->type])) {
				it->statLife = 0;
			}
		}
	}

	// 8-9| Velocity correction and integration
	i = -1;
	for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
		// 8| Velocity correction
		i++;
		bubbleVelocityCorrection(parameters.L[1], &(*it), stopFlag);
	}
	for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
		// 9| Velocity integration
		it->x = it->x + it->ux*dt;
		it->y = it->y + it->uy*dt;
		double velocity = sqrt(SQ(it->ux) + SQ(it->uy));
		if (velocity > Bubbles.maxVelocity) Bubbles.maxVelocity = velocity;
		// Check for clogging
		if ((it->ux == 0) && (it->uy == 0)) it->tClogg += dt;
		else it->tClogg = 0;

		if (it->tClogg > parameters.tClogg) {
			errorProgress("clogging detected @t=" + std::to_string(t) + "s");
			*stopFlag = true;
			return;
		}
		// Update wall interactions for next step (detached bubble growing)
		checkWalls(it->interactor.walls, Walls, it->x, it->y, 0.5*interactionLongMaximumRange*it->r);
	}

	// ?| Detached bubble growing 	/* !!!!!!! DISABLED (\/ see end of this file \/) !!!!!!! */

	// 10| Attached bubble growing
	for (it = Bubbles.attached.begin(); it != Bubbles.attached.end(); it++) {
		// Prise en compte du dr propre au site
		Site* site = it->site;
		double dr = MIN(site->growth*(dt+site->tmiss), MAX(0, RFritzExtra[site->type] - it->r));
		site->tmiss = 0;

		double maxDr = INF;
		double itx = it->x + interactionSafetyHalfRange*dr*site->nx, ity = it->y + interactionSafetyHalfRange*dr*site->ny;

		// Detection des parois ...
		checkWalls(it->interactor.walls, Walls, itx, ity, 0.5*interactionLongMaximumRange*it->r);
		for (int kk = 0; kk < wallMaxmimumIndicators; kk++) {
			WallIndicator *indicator = &it->interactor.walls[kk];
			if (indicator->distance < INF) {
				double deltaR = (it->r+dr);
				if (SQ(indicator->distance) < interactionSafetyHalfRangeSq*SQ(deltaR)) {
					double drRemaining = interactionSafetyHalfRangeInv*indicator->distance - it->r;
					if (drRemaining < maxDr) {
						maxDr = drRemaining;
					}
				}
			}
		}
		// ... et des bulles empechant la croissance
		for (int kk = 0; kk < it->interactor.count; kk++) {
			Bubble* kt = it->interactor.bubbles[kk];
			int kMode = it->interactor.imagedModes[kk];
			double deltaX = (kt->x - itx), deltaY = (yRepeat(kt->y, parameters.L[1], kMode) - ity), deltaR = (it->r+dr + kt->r);
			double dSQ = deltaX*deltaX + deltaY*deltaY;
			if (dSQ < interactionSafetyHalfRangeSq*SQ(deltaR)) {
				double drRemaining = interactionSafetyHalfRangeInv*sqrt(dSQ) - it->r - kt->r;
				if (drRemaining < maxDr) {
					maxDr = drRemaining;
				}
			}
		}

		// Si le grossissement a été limitée, il faut le reporter
		if (maxDr != INF) {
			double drBis = MIN(MAX(0, maxDr), dr);
			site->tmiss = (dr-drBis)/site->growth;
			dr = drBis;
		}

		// La croissance ne peut commencer qu'à partir d'un rayon mini
		if (it->r == 0) {
			if ((it->r+dr) < Rmini[site->type]) {
				site->tmiss += dr/site->growth;
				dr = 0;
			}
		}

		// Finalement, si grossissement il y a :
		if (dr != 0) {
			it->quantity[site->type] += nRPrefactor*(mathCB(it->r + dr)-mathCB(it->r));
			it->r = it->r + dr;
			it->x = it->x + interactionSafetyHalfRange*dr*site->nx;
			it->y = it->y + interactionSafetyHalfRange*dr*site->ny;
		}
	}

	// 11| Detached bubble exit
	list<list<Bubble>::iterator> toKillDetached;
	for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
		if (it->x < 0.0) { // left exit (0)
			it->statLife = 0;
			// Add to stats
			Statistics.collected[0] += it->quantity[0];
			Statistics.contamination[0] += it->quantity[1];
			Statistics.lost[1] += it->quantity[1];
		} else if (it->x > parameters.L[0]) { // right exit (1)
			it->statLife = 0;
			// Add to stats
			Statistics.collected[1] += it->quantity[1];
			Statistics.contamination[1] += it->quantity[0];
			Statistics.lost[0] += it->quantity[0];
		} else if (it->y > parameters.L[1]) { // top recopy
			it->y = it->y - parameters.L[1];
			it->statLife = it->statLife - 1;
			// Add to stats
			if (it->statLife == 0) {
				Statistics.lost[0] += it->quantity[0];
				Statistics.lost[1] += it->quantity[1];
			}
		}
		if (it->statLife == 0) {
			Bubbles.countDetached--;
			Bubbles.flagDetached = true;
			toKillDetached.push_front(it);
			continue;
		}
	}
	if (Bubbles.flagDetached) { // If a bubble dies, reorder the list
		for (list<list<Bubble>::iterator>::iterator itKill = toKillDetached.begin(); itKill != toKillDetached.end(); itKill++) Bubbles.detached.erase(*itKill);
		Bubbles.flagDetached = false;
	}

	// 12| Attached bubble detachment
	list<list<Bubble>::iterator> toKillAttached;
	for (it = Bubbles.attached.begin(); it != Bubbles.attached.end(); it++) {
		if (it->statLife == 0) {
			// Changement de statut
			Bubbles.flagAttached = true;
			it->statLife = bubbleLifeCount;
			toKillAttached.push_front(it);
			Bubbles.detached.push_back(*it);
			Bubbles.countAttached--; Bubbles.countDetached++;
			// Prochaine nucléation
			it->site->t = t + growingKineticFractionWait*it->site->T;
			it->site->occuped = false;
		}
	}

	// 13| Reorder bubble list
	if (Bubbles.flagAttached) { // If a bubble dies, reorder the list
		for (list<list<Bubble>::iterator>::iterator itKill = toKillAttached.begin(); itKill != toKillAttached.end(); itKill++) Bubbles.attached.erase(*itKill);
		Bubbles.flagAttached = false;
	}
}
/* ------------------------------ */






// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! DISABLED !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// ?| Detached bubble growing
/*double dr = dt*growingFactor;
int growingBubbleCount; Bubble* growingCollisionBubbles[2]; int growingCollisionBubblesMode[2]; double growingCollisionBubblesD[2];
WallIndicator* growingCollisionWall = 0; int growingWallCount;
for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
	growingBubbleCount = 0; growingWallCount = 0;
	// Detection des bulles empechant la croissance
	for (int kk = 0; kk < it->interactor.count; kk++) {
		if (it->interactor.distancesSqNorm[kk] < interactionCollisionRangeSq) {
			Bubble* kt = it->interactor.bubbles[kk];
			int kMode = it->interactor.imagedModes[kk];
			double deltaX = (kt->x - it->x), deltaY = (Y_repeat(kt->y, kMode) - it->y), deltaR = (it->r+dr + kt->r);
			double dSQ = deltaX*deltaX + deltaY*deltaY;
			if (dSQ < interactionSafetyHalfRangeSq*SQ(deltaR)) {
				if (growingBubbleCount == 2) break;
				growingCollisionBubbles[growingBubbleCount] = kt;
				growingCollisionBubblesD[growingBubbleCount] = sqrt(dSQ);
				growingCollisionBubblesMode[growingBubbleCount] = kMode;
				growingBubbleCount++;
			}
		}
	}
	// Detection des parois empechant la croissance
	for (int kk = 0; kk < wallMaxmimumIndicators; kk++) {
		WallIndicator *indicator = &it->interactor.walls[kk];
		if (indicator->distance < (0.5*it->r*interactionCollisionRange)) {
			// Recalcul de la distance exact (à cause de l'intégration)
			if (indicator->distance < interactionSafetyHalfRange*(it->r+dr)) {
				if (growingWallCount == 1) break;
				growingCollisionWall = indicator;
				growingWallCount++;
			}
		}
	}
	if ((growingBubbleCount != 0) || (growingWallCount != 0)) {
		// Simple decallage si une bulle en blocage avec une autre
		if ((growingBubbleCount == 1) && (growingWallCount == 0)) {
			double nx = it->x - growingCollisionBubbles[0]->x, ny = it->y - Y_repeat(growingCollisionBubbles[0]->y, growingCollisionBubblesMode[0]);
			double d_inv = 1.0/growingCollisionBubblesD[0];
			nx = d_inv*nx; ny = d_inv*ny;

			it->x = it->x + dr*nx;
			it->y = it->y + dr*ny;
		// Idem avec une paroi
		} else if ((growingBubbleCount == 0) && (growingWallCount == 1)) {
			it->x = it->x + dr*growingCollisionWall->n.x;
			it->y = it->y + dr*growingCollisionWall->n.y;
		// Résolution de la position si deux bulles
		} else if ((growingBubbleCount == 2) && (growingWallCount == 0)) {
			continue;
		// Ou une bulle et une paroi
		} else if ((growingBubbleCount == 1) && (growingWallCount == 1)) {
			continue;
		// Pas de croissance si plus de deux bulles en blocage ou une bulle et un mur
		} else {
			continue;
		}
	}

	// Croissance
	it->r = it->r + dr;
	double quantity = quantityFromDr(it->r, it->r + dr), tension = getTension(parameters, TENSION, it->x, it->y);
	it->quantity[O2] += tension*quantity; it->quantity[H2] += (1-tension)*quantity;
	if (it->r > Bubbles.maxRadius)
		Bubbles.maxRadius = it->r;
}*/ // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! DISABLED !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
