// This file is part of BUDEES, a software that allows fast simulation of bubble dynamics in electrolyzers and their interactions with electrical parameters. 
// 
// Copyright (C) 2024 <CNRS-ENS Rennes-Université de Rennes-CY Cergy Paris Université-Université Paris Saclay-CNAM Paris-ENS Paris-Saclay>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/



// Compute total current (at the anode)
double computeTotalCurrent(Parameters& parameters, MatrixXi& ID, vector<ElectrodeNode>& nodes, VectorXd& phi_n, double phi_anode) {
	double I = 0;
	vector<ElectrodeNode>::iterator it;
	for (it = nodes.begin(); it != nodes.end(); it++) {
		double phi_n_id = phi_n(it->id);
		if (it->nx_anode != 0) I += it->nx_anode*eThick*parameters.dL[1]*TafelAnodeJ(phi_anode-phi_n_id);
		if (it->ny_anode != 0) I += it->ny_anode*eThick*parameters.dL[0]*TafelAnodeJ(phi_anode-phi_n_id);
	}
	return I;
}

// Non linear part of potential equation (Tafel)
VectorXd evaluateF(Parameters& parameters, MatrixXi& ID, vector<ElectrodeNode>& nodes, VectorXd& phi_n, double phi_anode, double phi_cathode) {
	double dxSigmaInv = 1.0/(sigma*parameters.dL[0]), dySigmaInv = 1.0/(sigma*parameters.dL[1]);

	VectorXd F_n(phi_n.size()); F_n.setZero();
	vector<ElectrodeNode>::iterator it;
	for (it = nodes.begin(); it != nodes.end(); it++) {
		int id = it->id;
		double phi_n_id = phi_n(it->id);
		if (it->nx_anode != 0) F_n(id) -= it->nx_anode*it->denomInv*dxSigmaInv*TafelAnodeJ(phi_anode-phi_n_id);
		if (it->ny_anode != 0) F_n(id) -= it->ny_anode*it->denomInv*dySigmaInv*TafelAnodeJ(phi_anode-phi_n_id);
		if (it->nx_cathode != 0) F_n(id) -= it->nx_cathode*it->denomInv*dxSigmaInv*TafelCathodeJ(phi_n_id-phi_cathode);
		if (it->ny_cathode != 0) F_n(id) -= it->ny_cathode*it->denomInv*dySigmaInv*TafelCathodeJ(phi_n_id-phi_cathode);
	}
	return F_n;
}
// Jacobian of the non linear part of potential equation (Tafel)
SpMat evaluateDF(Parameters& parameters, MatrixXi& ID, vector<ElectrodeNode>& nodes, VectorXd& phi_n, double phi_anode, double phi_cathode) {
	double dxSigmaInv = 1.0/(sigma*parameters.dL[0]), dySigmaInv = 1.0/(sigma*parameters.dL[1]);

	SpMat dF_n(phi_n.size(), phi_n.size());
	vector<Trip> coeffs;

	vector<ElectrodeNode>::iterator it;
	for (it = nodes.begin(); it != nodes.end(); it++) {
		double phi_n_id = phi_n(it->id), dF_n_id = 0;
		if (it->nx_anode != 0) dF_n_id -= it->nx_anode*it->denomInv*dxSigmaInv*dTafelAnodeJ(phi_anode-phi_n_id);
		if (it->ny_anode != 0) dF_n_id -= it->ny_anode*it->denomInv*dySigmaInv*dTafelAnodeJ(phi_anode-phi_n_id);
		if (it->nx_cathode != 0) dF_n_id -= it->nx_cathode*it->denomInv*dxSigmaInv*dTafelCathodeJ(phi_n_id-phi_cathode);
		if (it->ny_cathode != 0) dF_n_id -= it->ny_cathode*it->denomInv*dySigmaInv*dTafelCathodeJ(phi_n_id-phi_cathode);

		if (dF_n_id > 0) coeffs.push_back(Trip(it->id, it->id, dF_n_id));
	}
	dF_n.setFromTriplets(coeffs.begin(), coeffs.end());
	return dF_n;
}

// Calcul le potentiel en chaque noeuds de la solution électrolytique
// Pour la condition : anode = +phi_anode [V], cathode = -phi_cathode [V]
VectorXd subSolveTension(Parameters& parameters, SpMat& A, MatrixXi& ID, vector<ElectrodeNode>& nodes, double phi_anode, double phi_cathode, int* failFlag) {
	VectorXd phi_n(A.cols()); phi_n.setZero();
	VectorXd residual, delta_phi;

	int iteration = 0;
	double res0 = -1;
	while (1) {
		// Newton-Rahpson solving
		VectorXd F_n = evaluateF(parameters, ID, nodes, phi_n, phi_anode, phi_cathode);
		SpMat dF_n = evaluateDF(parameters, ID, nodes, phi_n, phi_anode, phi_cathode);
		residual = -(A*phi_n + F_n);

		// Compute and check residual
		double res = residual.norm()/phi_anode;
		if (res0 == -1) res0 = res;
		progressRefresh2a();
		if (res < TOLERANCE_POTENTIAL) break;

		// Start new iteration
		if (iteration > SOLVER_MAX_ITERATION) {
			*failFlag = -1;
			return {};
		} iteration++;

		// Next step
		SpMat Jacobian = A + dF_n;
		*failFlag = solveLU(delta_phi, Jacobian, residual);
		if (*failFlag != 0) return {};

		phi_n = phi_n + delta_phi;
	}
	return phi_n;
}

// Calcul le potentiel en chaque noeuds de la solution électrolytique
// En respectant la condition : int(J.dS) = parameters.I
MatrixTension solveTension(Parameters& parameters, MatrixXi& MASK, bool ignoreFail) {
	int* N = parameters.N;

	// Parcours de cellules du mask et attribution d'id seulement dans l'électrolyte
	MatrixXi ID(N[0], N[1]); ID.setZero();
	int n = 0;
	for (int i = 0; i < N[0]; i++) for (int j = 0; j < N[1]; j++) {
		if (MASK(i, j) == 0) ID(i, j) = n++;
	}
	double ddxInv = 1.0/mathSQ(parameters.dL[0]), ddyInv = 1.0/mathSQ(parameters.dL[1]);

	// Construction de la matrice des résistance (partie linéaire) et récupération des indices à l'interface electrolyte/electrode
	vector<Trip> coeffs;
	vector<ElectrodeNode> nodes;
	for (int i = 0; i < N[0]; i++) for (int j = 0; j < N[1]; j++) if (MASK(i,j) == 0) {
		int id = ID(i,j);
		ElectrodeNode node = {id, 0, 0, 0, 0, 0};
		bool electrolyteNeighboor[4] = {false, false, false, false};
		double denom = 0;

		// i-1
		if (i > 0) {
			if (MASK(i-1, j) == 0) {
				electrolyteNeighboor[0] = true;
				denom += ddxInv;
			} else if (MASK(i-1, j) == 1) node.nx_anode++;
			else if (MASK(i-1, j) == -1) node.nx_cathode++;
		}

		// i+1
		if (i < (N[0]-1)) {
			if (MASK(i+1, j) == 0) {
				electrolyteNeighboor[1] = true;
				denom += ddxInv;
			} else if (MASK(i+1, j) == 1) node.nx_anode++;
			else if (MASK(i+1, j) == -1) node.nx_cathode++;
		}

		// j-1 (with repeat condition)
		int jm = (j > 0) ? j-1 : (N[1]-1);
		if (MASK(i, jm) == 0) {
			electrolyteNeighboor[2] = true;
			denom += ddyInv;
		} else if (MASK(i, jm) == 1) node.ny_anode++;
		else if (MASK(i, jm) == -1) node.ny_cathode++;

		// j+1 (with repeat condition)
		int jp = (j < (N[1]-1)) ? j+1 : 0;
		if (MASK(i, jp) == 0) {
			electrolyteNeighboor[3] = true;
			denom += ddyInv;
		} else if (MASK(i, jp) == 1) node.ny_anode++;
		else if (MASK(i, jp) == -1) node.ny_cathode++;

		// i,j
		coeffs.push_back(Trip(id, id, 1));
		double denomInv = 1.0/denom;
		if (electrolyteNeighboor[0]) coeffs.push_back(Trip(id, ID(i-1,j), -ddxInv*denomInv));
		if (electrolyteNeighboor[1]) coeffs.push_back(Trip(id, ID(i+1,j), -ddxInv*denomInv));
		if (electrolyteNeighboor[2]) coeffs.push_back(Trip(id, ID(i,jm), -ddyInv*denomInv));
		if (electrolyteNeighboor[3]) coeffs.push_back(Trip(id, ID(i,jp), -ddyInv*denomInv));

		if ((node.nx_anode != 0) || (node.ny_anode != 0) || (node.nx_cathode != 0) || (node.ny_cathode != 0)) {
			// Un noeud ne peut être en contact à la fois avec l'anode et la cathode
			if (((node.nx_anode != 0) || (node.ny_anode != 0)) && ((node.nx_cathode != 0) || (node.ny_cathode != 0))) {
				errorProgress("electrical mesh too high (node in contact with both anode and cathode)");
				exit(1);
			}
			// Stockage du noeud frontière pour la partie non linéaire (Tafel)
			node.denomInv = denomInv;
			nodes.push_back(node);
		}
	}

	// Build linear matrix of resistors
	SpMat A(n, n);
	A.setFromTriplets(coeffs.begin(), coeffs.end());

	// Newton-Rahpson iterative solve
	int failFlag = 0;
	// First voltage approximation
	double phi_anode_cathode = parameters.J*parameters.L[0]/sigma + tafelA0[ANODE]*log(parameters.J/tafelJ0[ANODE]) + tafelA0[CATHODE]*log(parameters.J/tafelJ0[CATHODE]);
	double d_phi_anode_cathode = DERIVATIVE_VARIATION*phi_anode_cathode;

	VectorXd phi;
	double I = 0;
	int iteration = 0;
	double res0 = -1;
	while (1) {
		// Compute potential and current for delta_phi
		phi = subSolveTension(parameters, A, ID, nodes, phi_anode_cathode, 0.0, &failFlag);
		if (failFlag != 0) break;
		I = computeTotalCurrent(parameters, ID, nodes, phi, phi_anode_cathode);

		// Check residual and post progress
		double res = mathAbs(I-parameters.I)/parameters.I;
		if (res0 == -1) res0 = res;
		progressRefresh2b(mathMinMax(log(res0/res)/log(res0/TOLERANCE_CURRENT), 0.0, 1.0));
		if (res < TOLERANCE_CURRENT) break;

		// Start new iteration
		if (iteration > SOLVER_MAX_ITERATION) {
			failFlag = -2;
			break;
		} iteration++;

		// Compute derivative
		phi = subSolveTension(parameters, A, ID, nodes, phi_anode_cathode-d_phi_anode_cathode, 0.0, &failFlag);
		if (failFlag != 0) break;
		double dI = I - computeTotalCurrent(parameters, ID, nodes, phi, phi_anode_cathode-d_phi_anode_cathode);

		// Next step (Newton)
		phi_anode_cathode = phi_anode_cathode - d_phi_anode_cathode*(I-parameters.I)/(dI);
	}

	// Return
	MatrixXd TENSION(N[0], N[1]);
	if (failFlag == 0) {
		// Solving success : extract result
		for (int i = 0; i < N[0]; i++) for (int j = 0; j < N[1]; j++) {
			if (MASK(i,j) == 0)	TENSION(i,j) = phi(ID(i,j));
			else if (MASK(i,j) == 1) TENSION(i,j) = phi.maxCoeff();
			else if (MASK(i,j) == -1) TENSION(i,j) = phi.minCoeff();
			else if (MASK(i,j) == -2) TENSION(i,j) = -1.0;
			else if (MASK(i,j) == -3) TENSION(i,j) = phi.maxCoeff();
			else if (MASK(i,j) == -4) TENSION(i,j) = phi.minCoeff();
		}
		// Compute resistor (for effective surface S = e x Ly)
		double J = I/(eThick*parameters.L[1]);
		double resistor = (phi.maxCoeff() - phi.minCoeff()) / J;

		return {TENSION, phi.minCoeff(), phi.maxCoeff(), phi_anode_cathode, 0, resistor, 0, true};
	} else if (ignoreFail) {
		return {TENSION, 0, 0, 0, 0, 0, 0, false};
	} else {
		// Solving fail : stop program
		errorProgress("electrical solving error (code: " + std::to_string(failFlag) + ")");
		exit(1);
	}
}
