// This file is part of BUDEES, a software that allows fast simulation of bubble dynamics in electrolyzers and their interactions with electrical parameters. 
// 
// Copyright (C) 2024 <CNRS-ENS Rennes-Université de Rennes-CY Cergy Paris Université-Université Paris Saclay-CNAM Paris-ENS Paris-Saclay>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/

// Delete output content
namespace fs = std::filesystem;
void outputDelete(const char *folder) {
	for (const auto& entry : fs::directory_iterator(folder)) {
		if (entry.is_regular_file()) {
			fs::remove(entry.path());
		} else if (entry.is_directory()) {
			outputDelete(entry.path().string().c_str());
			fs::remove(entry.path());
		}
	}
}
// Create folder before calling loadBMP API
unsigned int dispatch_loadbmp_encode_file(const char *filename, const unsigned char *imageData, unsigned int width, unsigned int height, unsigned int components) {
	fs::path filePath(filename);
	string parentFolder = filePath.parent_path().string();
	if (!fs::exists(parentFolder)) fs::create_directory(parentFolder);

	return loadbmp_encode_file(filename, imageData, width, height, components);
}



// Text output
void outputData(const char *filename, Parameters& parameters, MatrixXd& MATRIX, double minVal, double maxVal) {
	ofstream frameFile;
	frameFile = ofstream(filename);
	if (!frameFile.is_open()) {
		cerr << "Unable to create outputfile." << endl;
		exit(1);
	}
	frameFile << "Data:" << endl;

	int width = MATRIX.rows(), height = MATRIX.cols();
	double x, y;
	for (int j = 0; j < (int) height; j++) for (int i = 0; i < (int) width; i++) {
		int jj = height - j - 1;
		indexToPosition(parameters.N, i, jj, &x, &y);
		double value = MATRIX(i,j);
		frameFile << parameters.L[0]*x << "," << parameters.L[1]*y << ",";
		if ((value == minVal) || (value == maxVal)) frameFile << "NaN";
		else frameFile << value;
		frameFile << endl;
	}
}
void outputData(string folder, const char *subfilename, Parameters& parameters, MatrixXd& MATRIX, double minVal, double maxVal) {
	string filename = folder + string(subfilename);
	outputData(filename.c_str(), parameters, MATRIX, minVal, maxVal);
}


// Bitmap output
void outputMatrix(const char *filename, MatrixXd& MATRIX, double minVal, double maxVal, double nullVal) {
	int width = MATRIX.rows(), height = MATRIX.cols();
	unsigned char *pixels = (unsigned char *) calloc(3*width*height, sizeof (unsigned char));
	unsigned int err;

	for (int i = 0; i < (int) width; i++) for (int j = 0; j < (int) height; j++) {
		int jj = height - j - 1;
		double v = (MATRIX(i,j) - minVal)/(maxVal-minVal);

		if (MATRIX(i,j) == nullVal) {
			pixels[3*(i+width*jj)+0] = 34;
			pixels[3*(i+width*jj)+1] = 177;
			pixels[3*(i+width*jj)+2] = 76;
		} else if (v <= 0.0) {
			pixels[3*(i+width*jj)+0] = 255;
			pixels[3*(i+width*jj)+1] = 255;
			pixels[3*(i+width*jj)+2] = 255;
		} else if (v >= 1.0) {
			pixels[3*(i+width*jj)+0] = 255;
			pixels[3*(i+width*jj)+1] = 255;
			pixels[3*(i+width*jj)+2] = 255;
		} else if ((v > 0) && (v < 1.0)) {
			pixels[3*(i+width*jj)+0] = mathMin(mathMax(0, mathMap(v, 0.5, 1.0, 0, 255)), 255);
			pixels[3*(i+width*jj)+2] = mathMin(mathMax(0, mathMap(v, 0.0, 0.5, 255, 0)), 255);

			for (int c = 0; c < 3; c++) {
				pixels[3*(i+width*jj)+c] = mathMin(mathMax(0, ((int) pixels[3*(i+width*jj)+c]) + ((int) mathMin(mathMax(0, 255-mathAbs(mathMap(v, 0.25, 0.75, -255, 255))), 255))), 255);
			}
		}
	}

	err = dispatch_loadbmp_encode_file(filename, pixels, width, height, LOADBMP_RGB);
	if (err) std::cout << "FAILURE: " << EXIT_FAILURE << std::endl;

	free(pixels);
}
void outputMatrix(string folder, const char *subfilename, MatrixXd& MATRIX, double minVal, double maxVal, double nullVal) {
	string filename = folder + string(subfilename);
	outputMatrix(filename.c_str(), MATRIX, minVal, maxVal, nullVal);
}
void outputMatrix(string folder, const char *subfolder, int i, MatrixXd& MATRIX, double minVal, double maxVal, double nullVal) {
	string filename = folder + string(subfolder) + string("frame_") + std::to_string(i) + string(".bmp");
	outputMatrix(filename.c_str(), MATRIX, minVal, maxVal, nullVal);
}


void outputMatrixI(const char *filename, MatrixXi& MATRIX, double minVal, double maxVal) {
	int width = MATRIX.rows(), height = MATRIX.cols();
	unsigned char *pixels = (unsigned char *) calloc(3*width*height, sizeof (unsigned char));
	unsigned int err;

	for (int i = 0; i < (int) width; i++) for (int j = 0; j < (int) height; j++) {
		int jj = height - j - 1;
		double v = (MATRIX(i,j) - minVal)/(maxVal-minVal);

		if (v == 0.0)
			pixels[3*(i+width*jj)+2] = 100;
		else if (v == 1.0)
			pixels[3*(i+width*jj)+0] = 100;
		else if ((v > 0) && (v < 1.0)) {
			pixels[3*(i+width*jj)+0] = MIN(MAX(0, 255.0*v), 255);
			pixels[3*(i+width*jj)+1] = MIN(MAX(0, 255.0*v), 255);
			pixels[3*(i+width*jj)+2] = MIN(MAX(0, 255.0*v), 255);
		} else {
			pixels[3*(i+width*jj)+0] = 255;
			pixels[3*(i+width*jj)+1] = 255;
			pixels[3*(i+width*jj)+2] = 255;
		}
	}
	err = dispatch_loadbmp_encode_file(filename, pixels, width, height, LOADBMP_RGB);
	if (err) std::cout << "FAILURE: " << EXIT_FAILURE << std::endl;

	free(pixels);
}
void outputMatrixI(string folder, const char *subfilename, MatrixXi& MATRIX, double minVal, double maxVal) {
	string filename = folder + string(subfilename);
	outputMatrixI(filename.c_str(), MATRIX, minVal, maxVal);
}

void outputMatrixI(string folder, const char *subfolder, int i, MatrixXi& MATRIX, double minVal, double maxVal) {
	string filename = folder + string(subfolder) + string("frame_") + std::to_string(i) + string(".bmp");
	outputMatrixI(filename.c_str(), MATRIX, minVal, maxVal);
}


void outputPoints(const char *filename, MatrixXd& POINTS, double normalizePoint, int RpxMax) {
	int width = POINTS.rows(), height = POINTS.cols();
	unsigned char *pixels = (unsigned char *) calloc(3*width*height, sizeof (unsigned char));
	unsigned int err;

	// Then add the points
	for (int i = 0; i < (int) width; i++) for (int j = 0; j < (int) height; j++) if (POINTS(i, j) != 0.0) {
		double v = MIN(MAX(-1.0, POINTS(i,j)/normalizePoint), 1.0);
		int Rpx = round(RpxMax*ABS(v));
		for (int ri = i-Rpx; ri <= i+Rpx; ri++) for (int rj = j-Rpx; rj <= j+Rpx; rj++) {
			int rjj = height - rj - 1;
			if ((ri > 0) && (ri < width-1) && (rj > 0) && (rj < height-1))
			if (SQ(ri-i) + SQ(rj-j) < SQ(Rpx)) {
				pixels[3*(ri+width*rjj)+0] = (v > 0.0) ? 255 : 0;
				pixels[3*(ri+width*rjj)+1] = 0;
				pixels[3*(ri+width*rjj)+2] = (v < 0.0) ? 255 : 0;
			}
		}
	}

	err = dispatch_loadbmp_encode_file(filename, pixels, width, height, LOADBMP_RGB);
	if (err) std::cout << "FAILURE: " << EXIT_FAILURE << std::endl;

	free(pixels);
}
void outputPoints(string folder, const char *subfilename, MatrixXd& POINTS, double normalizePoint, int RpxMax) {
	string filename = folder + string(subfilename);
	outputPoints(filename.c_str(), POINTS, normalizePoint, RpxMax);
}
void outputPoints(string folder, const char *subfolder, int i, MatrixXd& POINTS, double normalizePoint, int RpxMax) {
	string filename = folder + string(subfolder) + string("frame_") + std::to_string(i) + string(".bmp");
	outputPoints(filename.c_str(), POINTS, normalizePoint, RpxMax);
}


void outputPointsI(const char *filename, MatrixXi& POINTS, double normalizePoint, int RpxMax) {
	int width = POINTS.rows(), height = POINTS.cols();
	unsigned char *pixels = (unsigned char *) calloc(3*width*height, sizeof (unsigned char));
	unsigned int err;

	// Then add the points
	for (int i = 0; i < (int) width; i++) for (int j = 0; j < (int) height; j++) if (POINTS(i, j) != 0.0) {
		double v = MIN(MAX(-1.0, POINTS(i,j)/normalizePoint), 1.0);
		int Rpx = round(RpxMax*ABS(v));
		for (int ri = i-Rpx; ri <= i+Rpx; ri++) for (int rj = j-Rpx; rj <= j+Rpx; rj++) {
			int rjj = height - rj - 1;
			if ((ri > 0) && (ri < width-1) && (rj > 0) && (rj < height-1))
			if ((ri-i)*(ri-i) + (rj-j)*(rj-j) < Rpx*Rpx) {
				pixels[3*(ri+width*rjj)+0] = (v > 0.0) ? 255 : 0;
				pixels[3*(ri+width*rjj)+1] = 0;
				pixels[3*(ri+width*rjj)+2] = (v < 0.0) ? 255 : 0;
			}
		}
	}
	err = dispatch_loadbmp_encode_file(filename, pixels, width, height, LOADBMP_RGB);
	if (err) std::cout << "FAILURE: " << EXIT_FAILURE << std::endl;

	free(pixels);
}


void outputSites(const char *filename, Parameters& parameters, vector<Site>& Sites) {
	MatrixXd SITES(parameters.N[0], parameters.N[1]);
	for (vector<Site>::iterator it = Sites.begin(); it != Sites.end(); ++it) {
		double xNorm = it->x / parameters.L[0], yNorm = it->y / parameters.L[1];
		int i = MIN(MAX(0, round(parameters.N[0]*xNorm)), parameters.N[0]-1),
			j = MIN(MAX(0, round(parameters.N[1]*yNorm)), parameters.N[1]-1);
		SITES(i,j) = (it->type == H2 ? -1 : 1) * (1/it->T);
	}
	outputPoints(filename, SITES, MAX(-SITES.minCoeff(), SITES.maxCoeff()), 14);
}
void outputSites(string folder, const char *subfilename, Parameters& parameters, vector<Site>& Sites) {
	string filename = folder + string(subfilename);
	outputSites(filename.c_str(), parameters, Sites);
}
void outputSites(string folder, const char *subfolder, int i, Parameters& parameters, vector<Site>& Sites) {
	string filename = folder + string(subfolder) + string("frame_") + std::to_string(i) + string(".bmp");
	outputSites(filename.c_str(), parameters, Sites);
}





/*
 * Frame saving for visualization
 */


ofstream frameFile;
bool frameFileInit = false;
double outputDt, outputNext;
void initOutputFrame(const char *filename, Parameters& parameters, TimeSteps& timesteps, double tf, Topology& topology) {
	frameFile = ofstream(filename);
	if (!frameFile.is_open()) {
		cerr << "Unable to create outputfile." << endl;
		exit(1);
	}
	frameFile << parameters.L[0] << " " << parameters.L[1] << " " << eThick << " " << timesteps.dtElec << " ";

	int polyPos = 0;
	for (int s = 0; s < 2; s++) {
		frameFile << topology.polygonCount[s] << " ";
		for (int i = 0; i <= topology.polygonCount[s]; i++)
			frameFile << topology.polygonIndex[s][i] << " ";
		int polygonPtCount = topology.polygonIndex[s][topology.polygonCount[s]];
		for (int i = 0; i < polygonPtCount; i++) {
			frameFile << topology.polygonDefinitions[polyPos].x << ":" << topology.polygonDefinitions[polyPos].y << " ";
			polyPos++;
		}
	}
	frameFile << "!" << endl;

	if (outputMode == 1) {
		outputDt = outputTimeFactor/outputFps;
		outputNext = 0.0;
	}
	frameFileInit = true;
}
void outputFrame(string folder, const char *filename, double t, Parameters& parameters, TimeSteps& timesteps, Topology& topology, ListBubbles& Bubbles, ElectrolysisStatistics& Statistics) {
	string path = folder + string(filename);

	// Check init
	if (!frameFileInit)	initOutputFrame(path.c_str(), parameters, timesteps, parameters.tf, topology);

	// Check if frame must be written
	if (outputMode == 1) {
		if (t < outputNext) return;
		outputNext += outputDt;
	} else if (outputMode == 2) {
		if (t < outputStart) return;
		if (t > outputStop) return;
	}

	// Write global data
	updateRemaining(Statistics, Bubbles);
	frameFile << t << " " << Statistics.resistor << ":" << Statistics.overpotential << " ";
	frameFile << Statistics.tension[0] << ":" << Statistics.tension[1] << ":" << Statistics.current[0] << ":" << Statistics.current[1] << " ";
	frameFile << Statistics.contamination[1] << ":" << Statistics.collected[1] << ":" << Statistics.lost[1] << ":" << Statistics.remaining[1] << " ";

	// Write all bubble
	list<Bubble>::iterator it;
	for (it = Bubbles.attached.begin(); it != Bubbles.attached.end(); it++) {
		double itx, ity, itr;
		double itz = eThick/2;
		getAttachedBubbleLocation(itx, ity, itr, it);
		frameFile << "0:" << itx << ":" << ity << ":" << itz << ":" << itr << ":" <<  it->quantity[0] << ":" << it->quantity[1] << " ";
	}
	for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
		frameFile << "1:" << it->x << ":" << it->y << ":" << (eThick/2) << ":" << it->r << ":" <<  it->quantity[0] << ":" << it->quantity[1] << " ";
	}
	frameFile << "!" << endl;
}
void endOutputFrame(ofstream frameFile) {
	frameFile.close();
}
